from pathlib import Path
from typing import NamedTuple
from loguru import logger
import polars as pl
from plyze.qoi.data.data import to_multi_data
from plyze.qoi.data.interfaces import CaseQOIandData
from plyze.qoi.registries.main import QOIRegistry as QR


class StandardData(NamedTuple):
    zonal: pl.DataFrame
    surface: pl.DataFrame


def get_zonal_qois(idf: Path, sql: Path):
    qois = [
        QR.temp,
        QR.vent_vol,
        QR.vent_heat_gain,
        QR.vent_heat_loss,
        QR.custom.net_vent_heat_gain,
        QR.mix_vol,
        QR.mix_heat_gain,
        QR.mix_heat_loss,
        QR.custom.net_mix_heat_gain,
        QR.custom.combined_volume,
        # TODO: latent heat gain (which is not even in output variable requests currently), net incoming flow over a zone, net outgoing flow over a zone, combined mixing heat loss and heat gain
    ]
    return to_multi_data(qois, idf, sql)


def get_surface_qois(idf: Path, sql: Path):
    qois = [QR.flow_out, QR.flow_in, QR.custom.net_out_flow]
    return to_multi_data(qois, idf, sql)


# TODO: environmental qois
# NOTE: environmental qois would only need to be recorded once, and can actually be taken directly from the EPW, which is a different process


def gather_standard_data(
    idf: Path,
    sql: Path,
):

    zonal_df = get_zonal_qois(idf, sql)
    surface_df = get_surface_qois(idf, sql)
    return StandardData(zonal_df, surface_df)


def consolidate_data(case_datas: list[CaseQOIandData]):
    # assuming these are of all the same "type" of dataframe, ie zonal qois, or surface_qois, etc..
    logger.debug(case_datas[0].dataframe.shape)

    df = pl.concat(
        [i.dataframe.with_columns(case_name=pl.lit(i.case_name)) for i in case_datas]
    )
    expected_num_rows = sum(
        [i.dataframe["space_names"].unique().len() for i in case_datas]
        * case_datas[0].dataframe["datetimes"].unique().len()
    )
    assert df.height == expected_num_rows
    return df



# each node knows its level
# metrics taken from Ostwald 2011

from plyze.jpg.interfaces import JPGraph
from utils4plans.lists import sort_and_group_objects_dict
from plyze.jpg.interfaces import JPGMetrics


def calculate_total_depth(G: JPGraph):
    # sort and group ndoes based on level
    nodes = G.jpnodes
    levels = sort_and_group_objects_dict(nodes, lambda x: x.data.level)

    total_depth = 0
    for level, nodes in levels.items():
        val = level * len(nodes)
        # logger.debug(f"level: {level}, n_nodes: {len(nodes)}")
        total_depth += val

    return total_depth


def calculate_mean_depth(G: JPGraph, total_depth: float):
    return total_depth / (
        G.num_nodes - 1
    )  # TODO: this should be -1 if the carrier is included as one of the nodes


def calculate_relative_asymmetry(G: JPGraph, mean_depth: float):
    return 2 * (mean_depth - 1) / (G.num_nodes - 2)


def calculate_control_value(G: JPGraph):
    # TODO: fix -> tests not passing...
    def calc_b_value(node: str):
        return len(list(G.neighbors(node)))

    def calc_a_value(node: str):
        nbs = G.neighbors(node)

        cv_a = sum(1 / calc_b_value(b) for b in nbs)

        return cv_a

    control_values = dict()

    for node in G.nodes:
        control_values[node] = calc_a_value(node)

    return control_values


def calculate_jpg_metrics(G: JPGraph):
    total_depth = calculate_total_depth(G)
    mean_depth = calculate_mean_depth(G, total_depth)
    relative_asymmetry = calculate_relative_asymmetry(G, mean_depth)
    # control_value = calculate_control_value(G)
    # TODO: the control value will be the max control value or a list.. / some special way of going into the dataframe if want all ..

    return JPGMetrics(
        graph_name=G.graph_name,
        total_depth=total_depth,
        mean_depth=mean_depth,
        relative_asymmetry=relative_asymmetry,
        # control_value=control_value,
    )



import polars as pl
from pathlib import Path
import altair as alt
from plyze.qoi.registry import QOIRegistry, QOIandData

from plyze.qoi.data import to_dataframe_with_spaces


def plot_setup(qoidata: QOIandData):
    assert isinstance(
        qoidata.dataframe, pl.DataFrame
    ), f"Expected a dataframe, instead got: {qoidata.dataframe}. "
    chart = (
        alt.Chart(
            qoidata.dataframe,
        )
        .mark_point(filled=True, size=200)
        .encode(
            x=alt.X("area:Q").title("Area [m]").scale(zero=False),
            y=alt.Y(f"{qoidata.qoi.nickname}:Q").title(None).scale(zero=False),
            color=alt.Color("display_name").scale(scheme="dark2").title("Rooms"),
        )
    )
    return chart


def facet(chart: alt.Chart, label: str):
    return chart.properties(width=200, height=100).facet(
        # TODO: put this in "Metrics" registry..
        column=alt.Column("monthdate(datetimes):T").title(None),
        row=alt.Row("hours(datetimes):T").title(label),
    )


def corr_plot(qoidata: QOIandData):
    return facet(plot_setup(qoidata), qoidata.qoi.label)


def surface_corr_plot(qoidata: QOIandData):
    chart = (
        plot_setup(qoidata)
        .mark_point(filled=False, size=200)
        .encode(
            x=alt.X("direction").title("Direction"),
            color=alt.Color("zone_display_name:N")
            .sort()
            .scale(scheme="dark2")
            .title("Rooms Corresponding to Surface"),
            shape=alt.Shape("type_"),
            fill=alt.condition(
                "datum.type_ == 'Window'",
                alt.Color("zone_display_name:N").sort().scale(scheme="dark2"),
                alt.value("transparent"),
            ),
        )
    )
    return facet(chart, qoidata.qoi.label)


def zone_qois(idf: Path, sql: Path):
    qois = [
        QOIRegistry.temp,
        QOIRegistry.vent_vol,
        QOIRegistry.vent_heat_gain,
        QOIRegistry.vent_heat_loss,
        QOIRegistry.mix_vol,
        QOIRegistry.mix_heat_gain,
        QOIRegistry.mix_heat_loss,
    ]
    charts = [corr_plot(to_dataframe_with_spaces(i, idf, sql)) for i in qois]
    chart = charts[0] | charts[1]
    for c in charts[2:]:
        chart |= c
    return chart


def surface_qois(idf: Path, sql: Path):
    qois = [
        QOIRegistry.flow_12,
        QOIRegistry.flow_21,
    ]

    charts = [surface_corr_plot(to_dataframe_with_spaces(i, idf, sql)) for i in qois]
    chart = charts[0] | charts[1]
    return chart



from plyze.qoi.xarray_helpers import find_drn_in_name
from utils4plans.lists import sort_and_group_objects
from pathlib import Path
from plan2eplus.results.sql import get_qoi
from plyze.qoi.registries.interfaces import CustomQOIComponents, CustomQOI


# @dataclass(frozen=True)
# class OutgoingFlow(GenericQOI):
#     name: str
#
#     def fx(self, sql_path: Path):
#         # arr_a = get_qoi(self.components.a, sql_path).data_arr
#         # need to combine all outgoing flows for a zone and change space names to be zones..
#         # so here would need the space idfs...
#         # this could potentially alter the dag...
#         pass


# OTHER FUNCTIONS --- > Modifying the xarray dataarray in some more complex way...
def get_zone_total_incoming_flow(sql_path: Path):
    f21 = get_qoi("AFN Linkage Node 2 to Node 1 Volume Flow Rate", sql_path).data_arr
    # don't want to rely on the IDF, so use the zone temperatures to get the zone names and group together..
    pass


def get_wind_pressure_unique_external_nodes(sql_path: Path):
    wind_pressure = get_qoi("AFN Node Wind Pressure", sql_path).data_arr

    grouped_space_names = sort_and_group_objects(
        wind_pressure.space_names.to_series().to_list(), lambda x: find_drn_in_name(x)
    )
    # just want the first from each group
    selected_space_names = [i[0] for i in grouped_space_names]
    # TODO: good opportunity to simplify the names..  to be the direction..
    return wind_pressure.sel(space_names=selected_space_names)


class CustomQOIRegistry:
    net_out_flow = CustomQOI(  # NOTE: this is really more of a dissipation metric..
        name="AFN Linkage Node 1 to Node 2 Net Volume Flow Rate ",
        components=CustomQOIComponents(
            "AFN Linkage Node 1 to Node 2 Volume Flow Rate",
            "AFN Linkage Node 2 to Node 1 Volume Flow Rate",
        ),
        nickname="net_out_flow",
        unit="m3/s",
        space_type="Surface",
        info="Net OUTgoing flow",
    )
    combined_volume = CustomQOI(
        name="AFN Combined Mixing and Ventilation Volume",
        components=CustomQOIComponents(
            "AFN Zone Mixing Volume",
            "AFN Zone Ventilation Volume",
        ),
        operation="+",
        nickname="combined_volume",
        unit="m3",
        space_type="Zone",
    )
    net_vent_heat_gain = CustomQOI(
        name="AFN Zone Ventilation Net Sensible Heat Gain Rate",
        components=CustomQOIComponents(
            "AFN Zone Ventilation Sensible Heat Gain Rate",
            "AFN Zone Ventilation Sensible Heat Loss Rate",
        ),
        nickname="net_vent_heat_gain",
        unit="W",
        space_type="Zone",
    )
    net_mix_heat_gain = CustomQOI(
        name="AFN Zone Mixing Net Sensible Heat Gain Rate",
        components=CustomQOIComponents(
            "AFN Zone Mixing Sensible Heat Gain Rate",
            "AFN Zone Mixing Sensible Heat Loss Rate",
        ),
        nickname="net_mix_heat_gain",
        unit="W",
        space_type="Zone",
    )
    unique_wind_pressure = CustomQOI(
        name="AFN Facade External Node Wind Pressure",
        nickname="unique_wind_pressure",
        unit="Pa",
        space_type="System",
        fx=get_wind_pressure_unique_external_nodes,
    )



from plyze.qoi.registries.interfaces import EpQOI
from plyze.qoi.registries.custom import CustomQOIRegistry


class QOIRegistry:
    custom = CustomQOIRegistry
    flow_out = EpQOI(
        "AFN Linkage Node 1 to Node 2 Volume Flow Rate",
        "flow_out",
        "m3/s",
        "Surface",
        info="~Outgoing flows~: Outdoor flows: from thermal zone to outdoors. Indoor flows: from thermal zone 'owning' the surface to its neighbor.",
    )
    flow_in = EpQOI(
        "AFN Linkage Node 2 to Node 1 Volume Flow Rate",
        "flow_in",
        "m3/s",
        "Surface",
        info="~Incoming flows~: From outdoors to the thermal zone, from neighbors sharing surface to THIS zone (which `owns` the surface)",
    )

    temp = EpQOI("Zone Mean Air Temperature", "temp", "C", "Zone")
    mix_vol = EpQOI("AFN Zone Mixing Volume", "mix_vol", "m3", "Zone")
    vent_vol = EpQOI("AFN Zone Ventilation Volume", "vent_vol", "m3", "Zone")

    mix_heat_gain = EpQOI(
        "AFN Zone Mixing Sensible Heat Gain Rate",
        "mix_heat_gain",
        "W",
        "Zone",
    )
    vent_heat_gain = EpQOI(
        "AFN Zone Ventilation Sensible Heat Gain Rate",
        "vent_heat_gain",
        "W",
        "Zone",
    )

    mix_heat_loss = EpQOI(
        "AFN Zone Mixing Sensible Heat Loss Rate",
        "mix_heat_loss",
        "W",
        "Zone",
    )
    vent_heat_loss = EpQOI(
        "AFN Zone Ventilation Sensible Heat Loss Rate",
        "vent_heat_loss",
        "W",
        "Zone",
    )
    wind_pressure = EpQOI(
        "AFN Node Wind Pressure",
        "wind_pressure",
        "Pa",
        "System",
    )



from pathlib import Path
from plan2eplus.ezcase.ez import EZ
from plan2eplus.visuals.base.base_plot import BasePlot


def plan_plot(idf: Path, show: bool = True):
    case = EZ(idf)
    bp = (
        BasePlot(case.objects.zones)
        .plot_zones()
        .plot_zone_names()
        .plot_subsurfaces_and_surfaces(
            case.objects.airflow_network,
            case.objects.airboundaries,
            case.objects.subsurfaces,
        )
        .plot_cardinal_names()
        .plot_connections(
            case.objects.airflow_network,
            case.objects.airboundaries,
            case.objects.subsurfaces,
        )
        #
    )
    if show:
        bp.show()



from itertools import product
from typing import Sequence
import xarray as xr
from pathlib import Path
from plyze.qoi.data.interfaces import QOIandData
from plyze.qoi.registries.interfaces import QOIType
from plyze.qoi.xarray_helpers import convert_xarray_to_polars, select_time
from datetime import datetime

from plyze.qoi.data.spaces import create_space_df


def select_custom_times(
    qoidata: QOIandData,
    year: int = 2017,
    month: int = 7,
    days: list[int] = [1],
    times: list[int] = [9, 12, 15, 18, 21, 0],
):
    datetimes = [
        datetime(year=year, month=month, day=i, hour=j) for i, j in product(days, times)
    ]
    # TODO: datetime object that can handle complex selections... and that can pass around.. definitely DONT want to be passing these about one by one
    # TODO: for reproducibility purposes this should be stated in the yaml
    arr = select_time(qoidata.original_arr, datetimes)
    qoidata.set_array(arr)
    return qoidata


def to_dataframe(qoidata: QOIandData):
    assert isinstance(qoidata.arr, xr.DataArray)
    df = convert_xarray_to_polars(qoidata.arr, qoidata.qoi.nickname)
    return df


def to_dataframe_with_spaces(qoi: QOIType, idf: Path, sql: Path):
    qoid = select_custom_times(QOIandData(qoi, sql))
    df = to_dataframe(qoid)
    space_df = create_space_df(idf)

    dd = df.join(space_df, on="space_names")
    qoid.set_dataframe(dd)
    return qoid


def to_multi_data(qois: Sequence[QOIType], idf: Path, sql: Path):

    def to_df(qoi: QOIType):
        qoid = select_custom_times(QOIandData(qoi, sql))
        return to_dataframe(qoid)

    # TODO: add utils4plans as set_unique function
    space_types = [i.space_type for i in qois]
    assert (
        len(set(space_types)) == 1
    ), f"Expected the space_types to all be the same, but got: {space_types}"

    dfs = [to_df(i) for i in qois]

    d0 = dfs[0]
    for df in dfs[1:]:
        d0 = d0.join(df, on=["datetimes", "space_names"])

    space_df = create_space_df(idf)
    return d0.join(space_df, on="space_names")



import altair as alt
import matplotlib.pyplot as plt
from cyclopts import App
from loguru import logger
from utils4plans.logconfig import logset

from plyze.examples.casedata import ex
from plyze.qoi.data.interfaces import CaseQOIandData
from plyze.qoi.data.outputs import consolidate_data, get_surface_qois
from plyze.qoi.plots.altair_helpers import AltairRenderers
from plyze.qoi.plots.theme import default_theme

app = App()


def keep():
    default_theme()
    logger.debug("")
    plt.plot()


### ----- DATA --------
@app.command()
def get():
    res = get_surface_qois(*ex)
    logger.debug(res)
    logger.debug(res.columns)


@app.command()
def cons():
    df1 = get_surface_qois(*ex)
    df2 = get_surface_qois(*ex)
    case_names = ["c1", "c2"]
    case_datas = [CaseQOIandData(case, df) for case, df in zip(case_names, [df1, df2])]
    df = consolidate_data(case_datas)
    logger.debug(df)


### ------- SINGLE PLOTS


# def plot_vol():
#     qoid = data_create()
#     logger.debug(qoid.dataframe)
#     chart = corr_plot(qoid)
#     chart.show()
#
#
# @app.command()
# def plot_surfs():
#     qoid = custom_qoi()  # data_create()
#     logger.debug(qoid.dataframe)
#
#     chart = surface_corr_plot(qoid)
#     chart.show()


### ------- MULTI PLOTS
# @app.command()
# def plot_vol_many():
#     c = zone_qois(ProjectPaths.sample_idf, ProjectPaths.sample_sql)
#     c.show()
#
#
# @app.command()
# def plot_surfs_many():
#     c = surface_qois(ProjectPaths.sample_idf, ProjectPaths.sample_sql)
#     c.show()


### ------- BIVAR PLORS

### ---- JPG ----


### --- CLUSTERING ----
# @app.command()
# def kn():
#     iris = load_iris(as_frame=True)
#     X = iris.data[["sepal length (cm)"]].to_numpy()
#     model, labels = fit_samples(X, 3)
#     logger.debug(model)
#     logger.debug(labels)
#     df = prep_cluster_df(X, labels, ["sepal len"])
#     logger.debug(df)
#     chart = show_clusters(df, "sepal len")
#     chart.show()

# model = fit_neighbors(X, 3)
# return model.predict()
# return show_neighbors_one(model, [[1]])


### ------- END COMMANDS ---------


def main():
    AltairRenderers.set_renderer()
    alt.theme.enable("default_theme")
    logset(to_stderr=True)
    app()


if __name__ == "__main__":
    main()



from datetime import datetime
from typing import Annotated
import polars as pl
from pathlib import Path

from plyze.jpg.interfaces import JPGMetrics, JPGraphModel
from plyze.jpg.metrics import calculate_jpg_metrics
from plyze.jpg.main import idf_to_jpgraph
from loguru import logger
from cyclopts import App, Parameter


jpg = App(name="jpg")


@jpg.command()
def create(
    graph_name: str, idf_path: Path, sql_path: Path, date_time: datetime, jpg_path: Path
):
    """
    Parameters
    ----------
    date_time: str
        Datetime format must be %Y-%m-%dT%H:%M:%S.
        See [cyclopts rules on coercing dates](https://cyclopts.readthedocs.io/en/latest/rules.html#datetime)
    """
    # TODO: register datetime as help!, not function description
    G = idf_to_jpgraph(graph_name, idf_path, sql_path, date_time)
    JPGraphModel.write(G, jpg_path)
    logger.success("Finished writing JPG")


@jpg.command()
def create_metrics(jpg_path: Path, metrics_path: Path):
    G = JPGraphModel.read(jpg_path)
    metrics = calculate_jpg_metrics(G)
    metrics.write(metrics_path)
    logger.success("Finished writing JPG metrics")


@jpg.command()
def consolidate(
    metrics_paths: Annotated[list[Path], Parameter(consume_multiple=True)],
    csv_path: Path,
):
    all_metrics = [JPGMetrics.read(i) for i in metrics_paths]
    df = pl.DataFrame(all_metrics)
    df.write_csv(csv_path)
    logger.success("Finished consolidating JPG metrics")



from plyze.paths import ProjectPaths
from plyze.qoi.plots import (
    corr_plot,
    surface_corr_plot,
    surface_qois,
    to_dataframe_with_spaces,
    zone_qois,
)
from plyze.qoi.registry import QOIType
from pathlib import Path

### ------- DESIGN METRICS  ---------------------- #####

### ------- QUANTITIES OF INTEREST ---------------------- #####
### ------- SINGLE PLOTS


def plot_vol(qoi: QOIType, idf_path: Path, sql_path: Path):
    qoid = to_dataframe_with_spaces(qoi, idf_path, sql_path)
    chart = corr_plot(qoid)
    chart.show()


def plot_surface(qoi: QOIType, idf_path: Path, sql_path: Path):

    qoid = to_dataframe_with_spaces(qoi, idf_path, sql_path)

    chart = surface_corr_plot(qoid)
    chart.show()


### ------- MULTI PLOTS


def plot_vol_many():
    c = zone_qois(ProjectPaths.sample_idf, ProjectPaths.sample_sql)
    c.show()


def plot_surf_many():
    c = surface_qois(ProjectPaths.sample_idf, ProjectPaths.sample_sql)
    c.show()



from plyze.jpg.interfaces import JPGraph, JPNodeData, JPNode, JPGMetrics


class VillaAlpha:
    graph_name = "ostwald11_villa_alpha"
    nodes = [
        JPNode(name="entry", data=JPNodeData(is_carrier=True, level=0)),
        JPNode(name="A", data=JPNodeData(level=1)),
        JPNode(name="B", data=JPNodeData(level=2)),
        JPNode(name="C", data=JPNodeData(level=5)),
        JPNode(name="D", data=JPNodeData(level=4)),
        JPNode(name="E", data=JPNodeData(level=3)),
        JPNode(name="F", data=JPNodeData(level=2)),
    ]

    edges = [
        ("entry", "A"),
        ("A", "F"),
        ("F", "E"),
        ("E", "D"),
        ("D", "C"),
    ]

    @property
    def graph(self):
        return JPGraph.create(self.graph_name, self.nodes, self.edges)

    @property
    def metrics(self):
        return JPGMetrics(
            graph_name=self.graph_name,
            total_depth=17,
            mean_depth=2.83,
            relative_asymmetry=0.73,
            control_value={"entry": 0.33, "A": 2.5, "B": 0.33, "F": 0.83},
        )



from sklearn.neighbors import NearestNeighbors
import altair as alt
import polars as pl
from sklearn.cluster import SpectralClustering
import numpy as np


# NEAREST NEIGBORS
def fit_neighbors(samples: np.ndarray, n: int):
    model = NearestNeighbors(n_neighbors=n)
    model.fit(samples)
    return model


def show_neighbors_one(model: NearestNeighbors, sample: np.ndarray):
    return model.kneighbors(sample)


# SPECTRAL CLUSTERING
def fit_samples(samples: np.ndarray, n_clusters: int):
    model = SpectralClustering(n_clusters=n_clusters).fit(samples)
    return model, model.labels_


def get_num_columns(arr: np.ndarray):
    return np.size(arr, axis=1)


def prep_cluster_df(samples: np.ndarray, labels: list[int], feature_names: list[str]):
    # assert get_num_columns(labels) == 1

    n_features = get_num_columns(samples)
    assert len(feature_names) == n_features

    df = pl.from_numpy(data=samples, schema=feature_names).with_columns(labels=labels)

    return df


def show_clusters(df: pl.DataFrame, feature: str):
    chart = (
        alt.Chart(df)
        .mark_point()
        .encode(x=alt.X(feature).scale(zero=False), color=alt.Color("labels:N"))
    ).properties(width=500, height=250)
    return chart

    # just 1d for now..
    #



from pathlib import Path
import pyprojroot


BASE_PATH = pyprojroot.find_root(pyprojroot.has_dir(".git"))


class StaticPaths:
    base = Path(BASE_PATH) / "static"
    inputs = base / "1_inputs"
    temp = base / "4_temp"


class ProjectPaths:
    sample_xarray = StaticPaths.inputs / "msd/data.nc"
    sample_sql = StaticPaths.inputs / "msd/eplusout.sql"
    sample_idf = StaticPaths.inputs / "msd/run.idf"



import polars as pl
from datetime import datetime
import xarray as xr
import re


def find_drn_in_name(space_name: str):
    pattern = re.compile("(NORTH)|(SOUTH)|(EAST)|(WEST)")
    res = pattern.search(space_name.upper())
    if res:
        return res.group()
    else:
        raise ValueError(
            f"External node name {space_name} does not contain a direction!"
        )


def select_time(arr: xr.DataArray, dt: datetime | list[datetime]):
    assert "datetimes" in arr.dims
    return arr.sel(datetimes=dt)


def convert_xarray_to_polars(data: xr.DataArray | xr.Dataset, name=""):
    if name:
        data.name = name
    return pl.from_pandas(data.to_dataframe(), include_index=True)


def get_data(arr: xr.DataArray):
    return arr.to_dict()["data"]



import altair as alt

CLEARVIEW = "ClearviewText"
FONT = f'{CLEARVIEW}, system-ui, -apple-system, BlinkMacSystemFont, ".SFNSText-Regular", sans-serif'
FONT_SIZE = 14
LABEL_FONT_SIZE = 14
TITLE_FONT_SIZE = 18
FONT_COLOR = "#161616"
LABEL_COLOR = "#525252"
DEF_WIDTH = 350


@alt.theme.register("default_theme", enable=True)
def default_theme() -> alt.theme.ThemeConfig:
    return {
        "config": {
            # "view": {
            #     "width": 350,
            #     "height": 280,
            # },
            "axis": {
                "labelColor": LABEL_COLOR,
                "labelFontSize": LABEL_FONT_SIZE,
                "labelFont": FONT,
                "labelFontWeight": 400,
                "titleColor": FONT_COLOR,
                "titleFontWeight": 400,
                "titleFontSize": TITLE_FONT_SIZE,
                "titleFont": FONT,
            },
            "axisX": {"titlePadding": 10},
            "axisY": {"titlePadding": 2.5},
            "text": {"font": FONT, "fontSize": FONT_SIZE},
            "range": {},  # type: ignore
            "legend": {
                "labelFont": FONT,
                "labelFontSize": LABEL_FONT_SIZE,
                "labelLimit": 500,
            },
            "header": {
                "labelFont": FONT,
                "titleFont": FONT,
                "labelFontSize": LABEL_FONT_SIZE,
                "titleFontSize": TITLE_FONT_SIZE,
            },
        },
    }



from pathlib import Path
from utils4plans.lists import get_unique_one
from datetime import datetime

from plan2eplus.ezcase.ez import EZ
from plan2eplus.ops.afn.ezobject import Airboundary
from plan2eplus.ops.subsurfaces.ezobject import Subsurface
from plan2eplus.ops.zones.ezobject import Zone
from plyze.jpg.interfaces import JPGraph, JPNode, JPNodeData
import networkx as nx
from utils4plans.sets import set_difference

# from plyze.qoi.custom_qois import find_drn
# from plyze.qoi.registry import QOIRegistry, QOIandData, select_time

# from plyze.qoi.registries.custom import find_drn_in_name

from plyze.qoi.xarray_helpers import find_drn_in_name, select_time
from plyze.qoi.registries.main import QOIRegistry
from plyze.qoi.data.interfaces import QOIandData


def set_levels(G: JPGraph):
    def update_level(node_name: str, level: int):
        node = G.get_jpnode_by_name(node_name)
        # NOTE: this is not very functional, should create a new data ..
        node.data.level = level
        G.update_jpnode(node.name, node.data)

    # assert carrier node has been set
    carrier_node = get_unique_one(G.jpnodes, lambda x: x.data.is_carrier)
    # set level of carrier node
    # update_level(carrier_node, 0)

    # set others based on distance from carrier..
    other_nodes = set_difference(G.nodes, [carrier_node.name])
    for node in other_nodes:
        try:
            distance = nx.shortest_path_length(G, source=carrier_node.name, target=node)
            # logger.debug(f"{carrier_node.name}, {node}, {distance}")
            update_level(node, int(distance))
        except nx.NetworkXNoPath:
            # TODO: the reason the metrics is failing is because maybe including invalid nodes in the count... maybe need a filtering function that takes only positive levels
            # NOTE: the node is disconnected, assign is a negative value for now, potentially delete later..
            update_level(node, -1)

    return G


def idf_to_jpgraph(
    graph_name: str, idf_path: Path, sql_path: Path, datetime_: datetime
):
    case = EZ(idf_path)

    def make_jpnode_from_zone(zone: Zone):
        return JPNode(name=zone.room_name, data=JPNodeData())

    def make_edge_from_surface(
        afn_surface: Subsurface | Airboundary, carrier_node_name: str
    ):
        node_names = [i.room_name for i in case.objects.zones] + [carrier_node_name]
        e = afn_surface.edge
        if e.space_a in node_names and e.space_b in node_names:
            return (e.space_a, e.space_b)

    def make_carrier_jpnode():
        # TODO: this should be pretty flexible, may want to make the carrier node be based on some other factor..
        wind_pressure_data = QOIandData(
            QOIRegistry.custom.unique_wind_pressure, sql_path
        ).original_arr
        max_node_at_time = (
            select_time(wind_pressure_data, datetime_).idxmax().to_dict()["data"]
        )
        max_node_as_drn = find_drn_in_name(max_node_at_time)
        return JPNode(name=max_node_as_drn, data=JPNodeData(is_carrier=True, level=0))

    carrier_node = make_carrier_jpnode()

    jpnodes = [make_jpnode_from_zone(i) for i in case.objects.zones] + [carrier_node]
    edges = [
        make_edge_from_surface(i, carrier_node.name)
        for i in case.objects.subsurfaces + case.objects.airboundaries
    ]
    filtered_edges = [i for i in edges if i]

    G = JPGraph.create(graph_name, jpnodes, filtered_edges)

    leveled_graph = set_levels(G)

    return leveled_graph



from dataclasses import dataclass
from typing import NamedTuple
import polars as pl
import xarray as xr
from pathlib import Path
from plan2eplus.results.sql import get_qoi
from plyze.qoi.registries.interfaces import EpQOI, CustomQOI, QOIType


@dataclass
class QOIandData:
    qoi: QOIType
    sql_path: Path
    arr: xr.DataArray | None = None
    dataframe: pl.DataFrame | None = None

    @property
    def original_arr(self):
        if isinstance(self.qoi, EpQOI):
            qoi_res = get_qoi(self.qoi.name, self.sql_path)
            # TODO: introduce some leniency for system variables.
            # assert (
            #     qoi_res.space_type == self.qoi.space_type
            # ), f"Expected space type is {qoi_res.space_type}, but plan2eplus space type is {self.qoi.space_type}"
            assert (
                qoi_res.unit == self.qoi.unit
            ), f"Expected unit for {self.qoi.name} is {qoi_res.unit}, but plan2eplus  unit is {self.qoi.unit}"
            return qoi_res.data_arr

        elif isinstance(self.qoi, CustomQOI):
            # TODO: may have a mapping? but can just change the fx attribute.. don't need a mapping..
            assert self.qoi.fx
            return self.qoi.fx(self.sql_path)
        raise ValueError(
            f"Can't compute arr unless type of self.qoi is CompQOI or EpQOI. Instead, type of variable {self.qoi.name} is {type(self.qoi)}"
        )

    def set_array(self, arr: xr.DataArray):
        self.arr = arr

    def set_dataframe(self, df: pl.DataFrame):
        self.dataframe = df


class CaseQOIandData(NamedTuple):
    case_name: str
    dataframe: pl.DataFrame

    @classmethod
    def read(cls, path: Path):
        case_name = pl.read_parquet_metadata(path)["case_name"]
        dataframe = pl.read_parquet(path)
        return cls(case_name, dataframe)

    def write(self, path):
        metadata = {"case_name": self.case_name}
        self.dataframe.write_parquet(path, metadata=metadata)



from dataclasses import dataclass
from typing import Literal
import xarray as xr
from pathlib import Path
from plan2eplus.results.sql import get_qoi

from functools import partial
from typing import Callable, NamedTuple
from plan2eplus.results.collections import SpaceTypesLiteral
from plan2eplus.ops.output.interfaces import OutputVariables


@dataclass(frozen=True)
class GenericQOI:
    name: str
    nickname: str
    unit: str
    space_type: SpaceTypesLiteral
    info: str = ""

    @property
    def label(self):
        return f"{self.name} [{self.unit}]"

    def update_xarray(self, arr: xr.DataArray):
        arr.name = self.name
        arr.attrs["units"] = self.unit
        # TODO would be good to relate to the info that Ladybug gives..
        # TODO: dont think are using the units on the actual xarray...


class CustomQOIComponents(NamedTuple):
    a: OutputVariables
    b: OutputVariables
    c: OutputVariables | None = None
    d: OutputVariables | None = None
    e: OutputVariables | None = None


def default_custom_qoi_fx(
    components: CustomQOIComponents, operation: Literal["+", "-"], sql_path: Path
) -> xr.DataArray:
    arr_a = get_qoi(components.a, sql_path).data_arr
    arr_b = get_qoi(components.b, sql_path).data_arr

    if operation == "-":
        return arr_a - arr_b
    elif operation == "+":
        return arr_a + arr_b


@dataclass(frozen=True)
class CustomQOI(GenericQOI):
    name: str
    components: CustomQOIComponents | None = None
    operation: Literal["+", "-"] = "-"
    fx: Callable[[Path], xr.DataArray] | None = None

    def __post_init__(self):
        if not self.fx:
            assert self.components
            assert self.operation
            f = partial(default_custom_qoi_fx, self.components, self.operation)
            object.__setattr__(self, "fx", f)


@dataclass(frozen=True)
class EpQOI(GenericQOI):
    name: OutputVariables


QOIType = CustomQOI | EpQOI



from dataclasses import dataclass
import polars as pl
from pathlib import Path
from plan2eplus.ezcase.ez import EZ
from plan2eplus.ops.subsurfaces.ezobject import Subsurface, SubsurfaceType
from plan2eplus.ops.zones.ezobject import Zone
from plan2eplus.results.collections import SpaceTypesLiteral
from utils4plans.lists import get_unique_one


@dataclass
class BasicSpaceData:
    idf_name: str
    display_name: str
    space_type: SpaceTypesLiteral


@dataclass
class BasicZoneData(BasicSpaceData):
    area: float


@dataclass
class BasicSurfaceData(BasicSpaceData):
    type_: SubsurfaceType
    direction: str
    zone_display_name: str
    area: float


def upper_idf_column(df: pl.DataFrame):
    return df.with_columns(space_names=pl.col("idf_name").str.to_uppercase())


def create_space_df(idf_path: Path):
    # TODO: enable passing in a graph so can look at centrality..
    case = EZ(idf_path)

    def make_zone(zone: Zone):
        return BasicZoneData(zone.zone_name, zone.room_name, "Zone", zone.domain.area)

    def make_subsurface(s: Subsurface):
        zone_name = s.surface.zone_name
        zone = get_unique_one(case.objects.zones, lambda x: x.zone_name == zone_name)
        return BasicSurfaceData(
            s.subsurface_name,
            s.display_name,
            "Surface",
            s.subsurface_type,
            s.surface.direction.name,
            zone.room_name,
            zone.domain.area,
        )

    zones = [make_zone(i) for i in case.objects.zones]
    subsurfaces = [make_subsurface(i) for i in case.objects.subsurfaces]
    zdf = pl.DataFrame(zones).pipe(upper_idf_column)
    sdf = pl.DataFrame(subsurfaces).pipe(upper_idf_column)
    df = pl.concat([zdf, sdf], how="diagonal")
    # TODO: split this for zone and surface spatial information ..
    return df



import polars as pl
from plyze.qoi.data import to_multi_data
from plyze.utils import CaseData

import altair as alt
from plyze.qoi.plots import facet
from plyze.qoi.registry import QOIRegistry, QOIType


def bivar_plot(df: pl.DataFrame, q1: QOIType, q2: QOIType):
    chart = (
        alt.Chart(df)
        .mark_point(filled=False, size=200)
        .encode(
            x=alt.X(f"{q1.nickname}:Q").title(q1.name).scale(zero=False),
            y=alt.Y(f"{q2.nickname}:Q").title(None).scale(zero=False),
            color=alt.Color("display_name").scale(scheme="dark2").title("Rooms"),
        )
    )
    return facet(chart, q2.name)


def multi_bivar_plot(cd: CaseData):
    qp1, qp2 = (
        [QOIRegistry.vent_vol, QOIRegistry.custom.net_vent_heat_gain],
        [
            QOIRegistry.mix_vol,
            QOIRegistry.custom.net_mix_heat_gain,
        ],
    )
    chart = alt.hconcat()
    for qp in [qp1, qp2]:
        df = to_multi_data(qp, *cd)
        chart |= bivar_plot(df, *qp)

    return chart



from typing import Hashable, Literal, NamedTuple, Sequence, TypeVar
from dataclasses import dataclass

import networkx as nx
import xarray as xr
from plan2eplus.geometry.coords import Coord

NodeType = Literal["zone", "external_node"]


class ZoneNodeData(NamedTuple):
    type_: NodeType
    location: Coord
    area: float
    aspect_ratio: float
    vent_volume: xr.DataArray
    mix_volume: xr.DataArray
    ach: xr.DataArray


class ExternalNodeData(NamedTuple):
    type_: NodeType
    location: Coord
    external_wind_pressure: xr.DataArray


@dataclass(frozen=True)
class AFNNode:
    name: str
    data: ExternalNodeData | ZoneNodeData

    @property
    def entry(self):
        return (self.name, {"data": self.data})


@dataclass(frozen=True)
class ZoneNode(AFNNode):
    name: str
    data: ZoneNodeData


@dataclass(frozen=True)
class ExternalNode(AFNNode):
    name: str
    data: ExternalNodeData


class AFNEdgeData(NamedTuple):
    net_flow_rate: xr.DataArray


class AFNEdge(NamedTuple):
    u: str
    v: str
    data: AFNEdgeData

    @property
    def entry(self):
        return (self.u, self.v, {"data": self.data})


class BasicEdge(NamedTuple):
    u: str
    v: str


AFNNodeType = TypeVar("AFNNodeType", bound=AFNNode)


class AFNGraph(nx.Graph):
    def add_afn_nodes(self, nodes: list[AFNNodeType]):
        self.add_nodes_from([i.entry for i in nodes])

    def add_afn_edges(self, edges: list[AFNEdge]):
        self.add_edges_from([i.entry for i in edges])

    @property
    def edges_with_data(self):
        edges = [AFNEdge(u, v, data["data"]) for u, v, data in self.edges(data=True)]
        return edges

    @property
    def zone_nodes(self):
        nodes = self.nodes(data=True)
        res = [
            ZoneNode(i, data["data"])
            for i, data in nodes
            if isinstance(data["data"], ZoneNodeData)
        ]
        return res

    @property
    def external_nodes(self) -> list[ExternalNode]:
        nodes = self.nodes(data=True)
        res = [
            ExternalNode(i, data["data"])
            for i, data in nodes
            if isinstance(data["data"], ExternalNodeData)
        ]
        return res

    @property
    def zone_names(self):
        return [i.name for i in self.zone_nodes]

    @property
    def external_node_names(self):
        return [i.name for i in self.external_nodes]

    @property
    def all_names(self):
        return self.nodes(data=False)

    @property
    def all_nodes(self):
        return self.zone_nodes + self.external_nodes

    @property
    def layout(self) -> dict[Hashable, tuple[float, float] | Sequence[float]]:
        return {node.name: list(node.data.location.as_tuple) for node in self.all_nodes}

    @property
    def zone_only_subgraph(self):
        sg = nx.Graph()
        sg.add_nodes_from(self.zone_names)
        return sg

    @property
    def external_node_only_subgraph(self):
        sg = nx.Graph()
        sg.add_nodes_from(self.external_node_names)
        return sg

    def make_time_specific_digraph(self, data: list[float]):
        # NOTE: this is a bit risky, becasue not sure the egdes and the the data have the same alignment... maybe should take in a list of the edges also.., but its a property, so should only be calculated once, then cached...
        def make_edge(edge: BasicEdge, datapt: float):
            if abs(datapt) > 0:
                return (edge.u, edge.v)
            else:
                return (edge.v, edge.u)

        G = nx.DiGraph()
        edges = [
            make_edge(BasicEdge(i.u, i.v), datapt)
            for i, datapt in zip(self.edges_with_data, data)
        ]

        G.add_edges_from([i for i in edges])
        return G



from utils4plans.logconfig import logset
import altair as alt

from cyclopts import App

from plyze.fpviz.main import plan_plot
from plyze.paths import ProjectPaths
from plyze.qoi.plots.altair_helpers import AltairRenderers
from plyze.qoi.plots.theme import default_theme
from loguru import logger
from plyze.cli.make.jpg import jpg
from plyze.cli.make.qoi import qoi

app = App()
app.command(jpg)
app.command(qoi)


def keep():
    default_theme()
    logger.debug("")


### ------- BEGIN COMMANDS ----------

### ------ SHOW FLOOR PLAN


@app.command()
def show_plan():
    plan_plot(ProjectPaths.sample_idf)


### ------- END COMMANDS ---------


def main():
    AltairRenderers.set_renderer()
    alt.theme.enable("default_theme")
    logset(to_stderr=True)
    app()


if __name__ == "__main__":
    main()



from pathlib import Path
import networkx as nx
from utils4plans.io import read_json, write_json
from utils4plans.lists import get_unique_one
from pydantic import BaseModel


# TODO: but do have some ability to control how this is read in.. when get to actually handlling io for the AFNGraph..
#


# @dataclass
class JPNodeData(BaseModel):
    is_carrier: bool = False
    level: int = 0
    # coord: Coord = Coord(0, 0)
    # ix_on_level: int = 0
    #


# @dataclass
class JPNode(BaseModel):
    name: str
    data: JPNodeData

    @property
    def entry(self):
        return (self.name, {"data": self.data})


class Edge(BaseModel):
    # TODO: include this in utils4plans, since all models for graphs involving IO will extend this
    source: str
    target: str

    @property
    def as_tuple(self):
        return (self.source, self.target)


class JPGraph(nx.Graph):
    graph_name: str = ""
    # TODO: init function that takes jpnodes and their edges..

    def add_jpnodes(self, nodes: list[JPNode]):
        self.add_nodes_from([i.entry for i in nodes])

    @classmethod
    def create(
        cls, name: str, nodes: list[JPNode], edges: list[Edge] | list[tuple[str, str]]
    ):
        G = cls()
        G.graph_name = name
        G.add_jpnodes(nodes)
        if isinstance(edges[0], Edge):
            # TODO: figure out how to assert for a group of edges
            G.add_edges_from(
                [
                    i.as_tuple  # pyright: ignore[reportAttributeAccessIssue]
                    for i in edges  # pyright: ignore[reportAttributeAccessIssue]
                ]
            )
        else:
            G.add_edges_from(edges)  # pyright: ignore[reportArgumentType]
        return G

    @property
    def num_nodes(self):
        return len(self)

    @property
    def jpnodes(self):
        nodes = self.nodes(data=True)
        res = [JPNode(name=i, data=data_["data"]) for i, data_ in nodes]
        return res

    @property
    def jpedges(self):
        edges = self.edges(data=False)
        res = [Edge(source=u, target=v) for u, v in edges]
        return res

    def update_jpnode(self, name: str, data: JPNodeData):
        # TODO: test
        nx.set_node_attributes(G=self, values={"data": data}, name=name)

    def get_jpnode_by_name(self, name: str):
        node = get_unique_one(self.jpnodes, lambda node: node.name == name)
        return node

    def show(self):
        levels = set([i.data.level for i in self.jpnodes])
        s = "\n"
        for level in levels:
            if level >= 0:
                nodes = [i.name for i in self.jpnodes if i.data.level == level]
                info = f"{level}: {",".join(nodes)}\n"
                s += info
        return s


class JPGraphModel(BaseModel):
    graph_name: str
    nodes: list[JPNode]
    edges: list[Edge]

    @classmethod
    def read(cls, path: Path):
        data = read_json(path)
        model = cls.model_validate(data)
        G = JPGraph.create(model.graph_name, model.nodes, model.edges)
        return G

    @classmethod
    def write(cls, G: JPGraph, path: Path):
        model = cls.model_validate(
            {"graph_name": G.graph_name, "nodes": G.jpnodes, "edges": G.jpedges}
        )
        write_json(model.model_dump(), path, OVERWRITE=True)


class JPGMetrics(BaseModel):
    graph_name: str
    total_depth: float
    mean_depth: float
    relative_asymmetry: float
    # control_value: dict[str, float]

    @classmethod
    def read(cls, path: Path):
        data = read_json(path)
        model = cls.model_validate(data)
        return model

    def write(self, path: Path):
        write_json(self.model_dump(), path, OVERWRITE=True)

        # TODO need to read for creating the unifying data...



from pathlib import Path

from loguru import logger

from plyze.qoi.data.interfaces import CaseQOIandData
from plyze.qoi.data.outputs import consolidate_data, gather_standard_data
from cyclopts import App, Parameter
from typing import Annotated

qoi = App(name="qoi")


@qoi.command()
def create(
    case_name: str, idf_path: Path, sql_path: Path, zonal_path: Path, surface_path: Path
):

    data = gather_standard_data(idf_path, sql_path)
    CaseQOIandData(case_name, data.zonal).write(zonal_path)
    CaseQOIandData(case_name, data.surface).write(surface_path)

    logger.success("Finished writing standard data")


@qoi.command()
def consolidate(
    in_paths: Annotated[list[Path], Parameter(consume_multiple=True)], out_path: Path
):
    datas = [CaseQOIandData.read(p) for p in in_paths]
    df = consolidate_data(datas)
    df.write_parquet(out_path)
    logger.success("Finished consolidating data")
